import asyncio
import tempfile
import time
import unittest
from unittest.mock import Mock, call
from datetime import datetime, UTC

import responses

from . import Alligater, DeferrableLogger, Feature, NoAssignment, Variant


class MockDeferredLogger(DeferrableLogger):
    def __init__(self):
        self.mock = Mock()

    def __call__(self, *args, **kwargs):
        self.mock(*args, **kwargs)

    def write_log(self, call_id, extra=None):
        self.mock.write_log(call_id, extra=extra)

    def drop_log(self, call_id):
        self.mock.drop_log(call_id)


class TestGater(unittest.IsolatedAsyncioTestCase):
    def test_no_features(self):
        """Instantiate with no features."""
        gater = Alligater()
        assert gater._features == {}

    async def test_no_yaml(self):
        """Instantiate with hardcoded features and no YAML."""
        logger = MockDeferredLogger()
        gater = Alligater(
            logger=logger,
            features=[
                Feature("foo", variants=[Variant("foo", "Foo")], default_arm="foo"),
            ],
        )
        assert await gater.foo({}) == "Foo"
        assert logger.mock.write_log.call_count == 2

    async def test_local_assignment_cache(self):
        def _sticky(feature, entity):
            raise NoAssignment

        foo = Feature(
            "foo",
            variants=[Variant("foo", "Foo"), Variant("bar", "Bar")],
            default_arm="foo",
        )

        mock_ts = datetime(2024, 1, 2, 3, 4, 5, tzinfo=UTC)
        mock_ts2 = datetime(2024, 1, 2, 3, 4, 6, tzinfo=UTC)
        gater = Alligater(features=[foo], sticky=_sticky, now=lambda: mock_ts2)
        gater._local_assignments.set(foo, {"id": "a"}, "bar", "Bar", mock_ts)
        assert await gater.foo({"id": "a"}) == "Bar"
        assert await gater.foo({"id": "b"}) == "Foo"
        assert gater._local_assignments.get(foo, {"id": "b"}) == (
            "foo",
            "Foo",
            mock_ts2,
        )
        gater._local_assignments.clear()
        assert await gater.foo({"id": "a"}) == "Foo"

    async def test_deferred_exposure_logging(self):
        def _sticky(feature, entity):
            if entity["id"] == 2:
                return "bar", "Bar", datetime(2024, 1, 2, 3, 4, 5, tzinfo=UTC)
            raise NoAssignment

        logger = MockDeferredLogger()
        gater = Alligater(
            logger=logger,
            sticky=_sticky,
            features=[
                Feature("foo", variants=[Variant("foo", "Foo")], default_arm="foo"),
            ],
        )
        v = await gater.foo({"id": 1}, deferred=True)
        assert v == "Foo"
        logger.mock.write_log.assert_has_calls(
            [call("e3e70682-c209-4cac-629f-6fbed82c07cd", extra=None)]
        )
        assert logger.mock.write_log.call_count == 1
        logger.mock.drop_log.assert_not_called()
        v.log()
        logger.mock.write_log.assert_has_calls(
            [call("e3e70682-c209-4cac-629f-6fbed82c07cd", extra=None)] * 2
        )
        logger.mock.drop_log.assert_not_called()

        logger.mock.reset_mock()
        v2 = await gater.foo({"id": 2}, deferred=True)
        assert v2 == "Bar"
        logger.mock.write_log.assert_not_called()
        logger.mock.drop_log.assert_not_called()
        call_id = v2._call_id
        v2.log()
        logger.mock.write_log.assert_called_with(call_id, extra=None)
        logger.mock.drop_log.assert_not_called()

    async def test_drop_deferred_logging(self):
        logger = MockDeferredLogger()
        gater = Alligater(
            sticky=lambda x, y: (
                "foo",
                "Foo",
                datetime(2024, 1, 2, 3, 4, 5, tzinfo=UTC),
            ),
            logger=logger,
            features=[
                Feature("foo", variants=[Variant("foo", "Foo")], default_arm="foo"),
            ],
        )
        v = await gater.foo({}, deferred=True)
        assert v == "Foo"
        call_id = v._call_id
        logger.mock.write_log.assert_not_called()
        logger.mock.drop_log.assert_not_called()
        del v
        logger.mock.write_log.assert_not_called()
        logger.mock.drop_log.assert_called_with(call_id)

    async def test_silence_logging(self):
        """Test that logging can be suppressed."""
        logger = Mock()
        gater = Alligater(
            logger=logger,
            features=[
                Feature("foo", variants=[Variant("foo", "Foo")], default_arm="foo"),
            ],
        )
        assert await gater.foo({}, silent=True) == "Foo"
        logger.assert_not_called()
        assert await gater.foo({}) == "Foo"
        logger.assert_called()

    async def test_yaml_no_reload(self):
        """Instantiate with YAML path and no reload interval."""
        with tempfile.NamedTemporaryFile() as tf:
            tf.write(
                b"""
feature:
  name: simplest_feature
  variants:
    foo: This is the only value
  default_arm: foo
"""
            )
            tf.flush()

            gater = Alligater(yaml=tf.name)

            assert await gater.simplest_feature({}) == "This is the only value"

    async def test_yaml_reload(self):
        """Instantiate with YAML path and reload interval."""
        with tempfile.NamedTemporaryFile() as tf:
            tf.write(
                b"""
feature:
  name: simplest_feature
  variants:
    foo: This is the only value
  default_arm: foo
"""
            )
            tf.flush()

            gater = Alligater(yaml=tf.name, reload_interval=1)
            assert await gater.simplest_feature({}) == "This is the only value"

            tf.write(
                b"""
feature:
  name: simplest_feature
  variants:
    foo: This is no longer the only value
    bar: This is another value
  default_arm: foo
  rollouts:
    - name: new_segment
      arms:
        - bar
      population:
        type: explicit
        value:
          - id_2

---
feature:
  name: another_feature
  variants:
    bar: new feature
  default_arm: bar
"""
            )
            tf.flush()

            time.sleep(2)
            assert (
                await gater.simplest_feature({}) == "This is no longer the only value"
            )
            assert (
                await gater.simplest_feature({"id": "id_2"}) == "This is another value"
            )
            assert await gater.another_feature({}) == "new feature"
            gater.stop()

    async def test_cross_ref_features(self):
        with tempfile.NamedTemporaryFile() as tf:
            tf.write(
                b"""
feature:
  name: ft1
  variants:
    var1: a
    var2: b
  rollouts:
    - name: segment1
      arms: [var2]
      population:
        type: explicit
        value: [id_2]
  default_arm: var1

---

feature:
  name: ft2
  variants:
    varX: x
    varY: y
  default_arm: varX
  rollouts:
    - name: segmentZ
      arms: [varY]
      population:
        type: feature
        name: ft1
        where: $value Eq 'b'

---

feature:
  name: ft3
  variants:
    varI: i
    varJ: j
  default_arm: varI
  rollouts:
    - name: segmentK
      arms: [varJ]
      population:
        type: feature
        name: ft1
        where: $variant Eq 'var2'
"""
            )
            tf.flush()

            gater = Alligater(yaml=tf.name, reload_interval=1)
            assert await gater.ft1({"id": "id_1"}) == "a"
            assert await gater.ft1({"id": "id_2"}) == "b"
            assert await gater.ft2({"id": "id_1"}) == "x"
            assert await gater.ft2({"id": "id_2"}) == "y"
            assert await gater.ft3({"id": "id_1"}) == "i"
            assert await gater.ft3({"id": "id_2"}) == "j"
            gater.stop()

    async def test_cross_ref_features_assigned_ts(self):
        with tempfile.NamedTemporaryFile() as tf:
            tf.write(
                b"""
feature:
  name: ft1
  variants:
    var1: a
    var2: b
  rollouts:
    - name: segment1
      arms: [var2]
      sticky: true
      population:
        type: explicit
        value: [id_2]
  default_arm: var1

---

feature:
  name: ft2
  variants:
    var_old: old
    var_new: new
  default_arm: var_old
  rollouts:
    - name: segmentK
      arms: [var_new]
      population:
        type: feature
        name: ft1
        where: TimeSince($assigned, 'mo') Lt 6
"""
            )
            tf.flush()

            mock_assigned_ts = datetime(2023, 1, 2, 3, 4, 5, tzinfo=UTC)
            mock_current_ts = datetime(2023, 8, 2, 3, 4, 6, tzinfo=UTC)

            def _sticky(feature, entity):
                if entity["id"] == "id_2" and feature.name == "ft1":
                    return "var2", "b", mock_assigned_ts
                raise NoAssignment

            gater = Alligater(
                yaml=tf.name,
                reload_interval=1,
                sticky=_sticky,
                now=lambda: mock_current_ts,
            )
            assert await gater.ft1({"id": "id_1"}) == "a"
            assert await gater.ft1({"id": "id_2"}) == "b"
            assert await gater.ft2({"id": "id_1"}) == "new"
            assert await gater.ft2({"id": "id_2"}) == "old"
            gater.stop()

    @responses.activate
    def test_yaml_remote_config(self):
        """Test with loading remote config."""
        url = "https://policylab.hks.harvard.edu/alligater/config.yaml"
        responses.add(
            responses.GET,
            url,
            body="""
feature:
  name: simplest_feature
  variants:
    foo: This is the only value
  default_arm: foo
""",
        )
        gater = Alligater(yaml=url)
        assert asyncio.run(gater.simplest_feature({})) == "This is the only value"



import asyncio
import copy
import json
import threading
import unittest
from dataclasses import dataclass
from datetime import datetime, timezone

import responses

from .arm import Arm
from .common import NoAssignment, SkipLog
from .feature import Feature
from .log import NetworkLogger, ObjectLogger
from .population import Population
from .rollout import Rollout
from .variant import Variant

fake_now = datetime(2022, 1, 29, 12, 11, 10, 0, tzinfo=timezone.utc)


def mock_now():
    return fake_now


mock_url = "https://test.glen/graphql"


@dataclass
class User:
    id: str


def wait_for_responses(logger, timeout=1.0, expected_count=1):
    """Test helper to wait for an NetworkLogger to write something."""
    d = timeout / 4.0
    t = 0.0
    with logger._cv:
        while True:
            assert t < timeout, "Timeout waiting for write"
            if expected_count == 0:
                logger._cv.wait(timeout=d)
                assert len(responses.calls) == 0, "Got unexpected write"
            if len(responses.calls) < expected_count:
                logger._cv.wait(timeout=d)
                t += d
            else:
                return


class MockWriter:
    def __init__(self):
        self.cv = threading.Condition()
        self.results = []

    def __call__(self, entity):
        with self.cv:
            self.results.append(entity)
            self.cv.notify_all()

    def assertWritten(self, expected, timeout=1.0):
        n = 0
        p = timeout / float(len(expected)) if expected else timeout
        with self.cv:
            while len(self.results) < len(expected) and n < timeout:
                self.cv.wait(timeout=p)
                n += p
            assert self.results == expected

    def assertNotWritten(self, timeout=1.0):
        with self.cv:
            self.cv.wait(timeout=timeout)
            assert not self.results


class TestObjectLogger(unittest.IsolatedAsyncioTestCase):
    async def test_log_simple(self):
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        write = MockWriter()
        v = await f(User("one"), log=ObjectLogger(write), now=mock_now)

        write.assertNotWritten()

        v.log()

        log1 = {
            "ts": fake_now,
            "call_id": "e3e70682-c209-4cac-629f-6fbed82c07cd",
            "entity": {
                "type": "User",
                "value": {
                    "id": "one",
                },
            },
            "feature": {
                "name": "test_feature",
                "rollouts": [
                    {
                        "arms": [
                            {
                                "type": "Arm",
                                "variant": "foo",
                                "weight": 1.0,
                            }
                        ],
                        "name": "default",
                        "population": {
                            "name": "Default",
                            "type": "Population",
                        },
                        "randomizer": "Hash(Concat('default', ':', $id))",
                        "sticky": None,
                        "type": "Rollout",
                    }
                ],
                "type": "Feature",
                "variants": {
                    "foo": {
                        "name": "foo",
                        "nested": False,
                        "type": "Variant",
                        "value": "Foo",
                    },
                },
            },
            "variant": {
                "name": "foo",
                "nested": False,
                "type": "Variant",
                "value": "Foo",
            },
            "assignment": "Foo",
            "repeat": False,
            "trace": None,
            "sticky": False,
        }

        write.assertWritten([log1])

        # Now log an exposure
        v.log()
        log2 = copy.deepcopy(log1)
        log2.update(
            {
                "call_id": log1["call_id"] + ":1",
                "repeat": True,
            }
        )

        write.assertWritten([log1, log2])

    async def test_log_simple_trace(self):
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        write = MockWriter()
        v = await f(User("two"), log=ObjectLogger(write, trace=True), now=mock_now)
        v.log()

        write.assertWritten(
            [
                {
                    "ts": fake_now,
                    "call_id": "e3e70682-c209-4cac-629f-6fbed82c07cd",
                    "entity": {
                        "type": "User",
                        "value": {
                            "id": "two",
                        },
                    },
                    "repeat": False,
                    "feature": {
                        "name": "test_feature",
                        "rollouts": [
                            {
                                "arms": [
                                    {
                                        "type": "Arm",
                                        "variant": "foo",
                                        "weight": 1.0,
                                    }
                                ],
                                "name": "default",
                                "population": {
                                    "name": "Default",
                                    "type": "Population",
                                },
                                "randomizer": "Hash(Concat('default', ':', $id))",
                                "sticky": None,
                                "type": "Rollout",
                            }
                        ],
                        "type": "Feature",
                        "variants": {
                            "foo": {
                                "name": "foo",
                                "nested": False,
                                "type": "Variant",
                                "value": "Foo",
                            },
                        },
                    },
                    "variant": {
                        "name": "foo",
                        "nested": False,
                        "type": "Variant",
                        "value": "Foo",
                    },
                    "assignment": "Foo",
                    "sticky": False,
                    "trace": [
                        {"type": "EnterGate", "data": {"feature": "test_feature"}},
                        {"type": "EnterFeature", "data": {"feature": "test_feature"}},
                        {"type": "EnterRollout", "data": {"rollout": "default"}},
                        {
                            "type": "EvaluatePopulation",
                            "data": {
                                "member": True,
                                "population": "Default",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "_Field",
                                "args": [{"id": "two"}, "$id"],
                                "result": "two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Concat",
                                "args": ["default", ":", "two"],
                                "result": "default:two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Hash",
                                "args": ["default:two"],
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "Randomize",
                            "data": {
                                "function": "Hash(Concat('default', ':', $id))",
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "EnterArm",
                            "data": {
                                "arm": {"type": "Arm", "variant": "foo", "weight": 1.0},
                                "cutoff": 1.0,
                                "x": 0.7560122309797355,
                            },
                        },
                        {"type": "LeaveArm", "data": {"matched": True}},
                        {"type": "LeaveRollout", "data": {"member": True}},
                        {
                            "type": "ChoseVariant",
                            "data": {"variant": "foo", "sticky": False},
                        },
                        {"type": "EnterVariant", "data": {"variant": "foo"}},
                        {"type": "LeaveVariant", "data": {"value": "Foo"}},
                        {"type": "LeaveFeature", "data": {"value": "Foo"}},
                        {"type": "LeaveGate", "data": {"value": "Foo"}},
                    ],
                }
            ]
        )

    async def test_log_sticky_trace(self):
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        mock_assign_ts = datetime(2022, 1, 29, 12, 11, 11, 0, tzinfo=timezone.utc)
        write = MockWriter()
        v = await f(
            User("two"),
            sticky=lambda f, e: ("stickyvariant", "sticky", mock_assign_ts),
            log=ObjectLogger(write, trace=True),
            now=mock_now,
        )
        v.log()

        write.assertWritten(
            [
                {
                    "ts": fake_now,
                    "call_id": "e3e70682-c209-4cac-629f-6fbed82c07cd:1",
                    "entity": {
                        "type": "User",
                        "value": {
                            "id": "two",
                        },
                    },
                    "feature": {
                        "name": "test_feature",
                        "rollouts": [
                            {
                                "arms": [
                                    {
                                        "type": "Arm",
                                        "variant": "foo",
                                        "weight": 1.0,
                                    }
                                ],
                                "name": "default",
                                "population": {
                                    "name": "Default",
                                    "type": "Population",
                                },
                                "randomizer": "Hash(Concat('default', ':', $id))",
                                "sticky": None,
                                "type": "Rollout",
                            }
                        ],
                        "type": "Feature",
                        "variants": {
                            "foo": {
                                "name": "foo",
                                "nested": False,
                                "type": "Variant",
                                "value": "Foo",
                            },
                        },
                    },
                    "repeat": True,
                    "sticky": True,
                    "variant": {"name": "stickyvariant"},
                    "assignment": "sticky",
                    "trace": [
                        {"type": "EnterGate", "data": {"feature": "test_feature"}},
                        {"type": "EnterFeature", "data": {"feature": "test_feature"}},
                        {
                            "type": "StickyAssignment",
                            "data": {
                                "variant": "stickyvariant",
                                "value": "sticky",
                                "assigned": True,
                                "ts": mock_assign_ts.isoformat(),
                                "source": "remote",
                            },
                        },
                        {"type": "LeaveFeature", "data": {"value": "sticky"}},
                        {"type": "LeaveGate", "data": {"value": "sticky"}},
                    ],
                }
            ]
        )

    async def test_log_sticky_trace_new(self):
        """Sticky function is passed but returns nothing in this case."""
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        write = MockWriter()

        def sticker(f, e):
            raise NoAssignment

        v = await f(
            User("two"),
            sticky=sticker,
            log=ObjectLogger(write, trace=True),
            now=mock_now,
        )
        v.log()

        write.assertWritten(
            [
                {
                    "ts": fake_now,
                    "call_id": "e3e70682-c209-4cac-629f-6fbed82c07cd",
                    "entity": {
                        "type": "User",
                        "value": {
                            "id": "two",
                        },
                    },
                    "feature": {
                        "name": "test_feature",
                        "rollouts": [
                            {
                                "arms": [
                                    {
                                        "type": "Arm",
                                        "variant": "foo",
                                        "weight": 1.0,
                                    }
                                ],
                                "name": "default",
                                "population": {
                                    "name": "Default",
                                    "type": "Population",
                                },
                                "randomizer": "Hash(Concat('default', ':', $id))",
                                "sticky": None,
                                "type": "Rollout",
                            }
                        ],
                        "type": "Feature",
                        "variants": {
                            "foo": {
                                "name": "foo",
                                "nested": False,
                                "type": "Variant",
                                "value": "Foo",
                            },
                        },
                    },
                    "repeat": False,
                    "sticky": True,
                    "variant": {
                        "name": "foo",
                        "nested": False,
                        "type": "Variant",
                        "value": "Foo",
                    },
                    "assignment": "Foo",
                    "trace": [
                        {"type": "EnterGate", "data": {"feature": "test_feature"}},
                        {"type": "EnterFeature", "data": {"feature": "test_feature"}},
                        {
                            "type": "StickyAssignment",
                            "data": {
                                "variant": None,
                                "value": None,
                                "assigned": False,
                                "ts": None,
                                "source": None,
                            },
                        },
                        {"type": "EnterRollout", "data": {"rollout": "default"}},
                        {
                            "type": "EvaluatePopulation",
                            "data": {
                                "member": True,
                                "population": "Default",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "_Field",
                                "args": [{"id": "two"}, "$id"],
                                "result": "two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Concat",
                                "args": ["default", ":", "two"],
                                "result": "default:two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Hash",
                                "args": ["default:two"],
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "Randomize",
                            "data": {
                                "function": "Hash(Concat('default', ':', $id))",
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "EnterArm",
                            "data": {
                                "arm": {"type": "Arm", "variant": "foo", "weight": 1.0},
                                "cutoff": 1.0,
                                "x": 0.7560122309797355,
                            },
                        },
                        {"type": "LeaveArm", "data": {"matched": True}},
                        {"type": "LeaveRollout", "data": {"member": True}},
                        {
                            "type": "ChoseVariant",
                            "data": {"variant": "foo", "sticky": True},
                        },
                        {"type": "EnterVariant", "data": {"variant": "foo"}},
                        {"type": "LeaveVariant", "data": {"value": "Foo"}},
                        {"type": "LeaveFeature", "data": {"value": "Foo"}},
                        {"type": "LeaveGate", "data": {"value": "Foo"}},
                    ],
                }
            ]
        )

    async def test_log_sticky_rollout_override_on(self):
        """Sticky function is not passed, but stipulated by rollout.

        This situation is a little bit fishy, but we technically allow it.
        """
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            rollouts=[
                Rollout(
                    sticky=True,
                    name=Rollout.DEFAULT,
                    population=Population.DEFAULT,
                    arms=[Arm("foo", weight=1.0)],
                )
            ],
        )

        write = MockWriter()

        v = await f(User("two"), log=ObjectLogger(write, trace=True), now=mock_now)
        v.log()

        write.assertWritten(
            [
                {
                    "ts": fake_now,
                    "call_id": "e3e70682-c209-4cac-629f-6fbed82c07cd",
                    "entity": {
                        "type": "User",
                        "value": {
                            "id": "two",
                        },
                    },
                    "feature": {
                        "name": "test_feature",
                        "rollouts": [
                            {
                                "arms": [
                                    {
                                        "type": "Arm",
                                        "variant": "foo",
                                        "weight": 1.0,
                                    }
                                ],
                                "name": "default",
                                "population": {
                                    "name": "Default",
                                    "type": "Population",
                                },
                                "randomizer": "Hash(Concat('default', ':', $id))",
                                "sticky": True,
                                "type": "Rollout",
                            }
                        ],
                        "type": "Feature",
                        "variants": {
                            "foo": {
                                "name": "foo",
                                "nested": False,
                                "type": "Variant",
                                "value": "Foo",
                            },
                        },
                    },
                    "repeat": False,
                    "sticky": True,
                    "variant": {
                        "name": "foo",
                        "nested": False,
                        "type": "Variant",
                        "value": "Foo",
                    },
                    "assignment": "Foo",
                    "trace": [
                        {"type": "EnterGate", "data": {"feature": "test_feature"}},
                        {"type": "EnterFeature", "data": {"feature": "test_feature"}},
                        {"type": "EnterRollout", "data": {"rollout": "default"}},
                        {
                            "type": "EvaluatePopulation",
                            "data": {
                                "member": True,
                                "population": "Default",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "_Field",
                                "args": [{"id": "two"}, "$id"],
                                "result": "two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Concat",
                                "args": ["default", ":", "two"],
                                "result": "default:two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Hash",
                                "args": ["default:two"],
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "Randomize",
                            "data": {
                                "function": "Hash(Concat('default', ':', $id))",
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "EnterArm",
                            "data": {
                                "arm": {"type": "Arm", "variant": "foo", "weight": 1.0},
                                "cutoff": 1.0,
                                "x": 0.7560122309797355,
                            },
                        },
                        {"type": "LeaveArm", "data": {"matched": True}},
                        {"type": "LeaveRollout", "data": {"member": True}},
                        {
                            "type": "ChoseVariant",
                            "data": {"variant": "foo", "sticky": True},
                        },
                        {"type": "EnterVariant", "data": {"variant": "foo"}},
                        {"type": "LeaveVariant", "data": {"value": "Foo"}},
                        {"type": "LeaveFeature", "data": {"value": "Foo"}},
                        {"type": "LeaveGate", "data": {"value": "Foo"}},
                    ],
                }
            ]
        )

    async def test_log_sticky_rollout_override_off(self):
        """Sticky function is passed, but rollout turns it off."""
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            rollouts=[
                Rollout(
                    sticky=False,
                    name=Rollout.DEFAULT,
                    population=Population.DEFAULT,
                    arms=[Arm("foo", weight=1.0)],
                )
            ],
        )

        write = MockWriter()

        def sticker(f, e):
            raise NoAssignment

        v = await f(
            User("two"),
            sticky=sticker,
            log=ObjectLogger(write, trace=True),
            now=mock_now,
        )
        v.log()

        write.assertWritten(
            [
                {
                    "ts": fake_now,
                    "call_id": "e3e70682-c209-4cac-629f-6fbed82c07cd",
                    "entity": {
                        "type": "User",
                        "value": {
                            "id": "two",
                        },
                    },
                    "feature": {
                        "name": "test_feature",
                        "rollouts": [
                            {
                                "arms": [
                                    {
                                        "type": "Arm",
                                        "variant": "foo",
                                        "weight": 1.0,
                                    }
                                ],
                                "name": "default",
                                "population": {
                                    "name": "Default",
                                    "type": "Population",
                                },
                                "randomizer": "Hash(Concat('default', ':', $id))",
                                "sticky": False,
                                "type": "Rollout",
                            }
                        ],
                        "type": "Feature",
                        "variants": {
                            "foo": {
                                "name": "foo",
                                "nested": False,
                                "type": "Variant",
                                "value": "Foo",
                            },
                        },
                    },
                    "repeat": False,
                    "sticky": False,
                    "variant": {
                        "name": "foo",
                        "nested": False,
                        "type": "Variant",
                        "value": "Foo",
                    },
                    "assignment": "Foo",
                    "trace": [
                        {"type": "EnterGate", "data": {"feature": "test_feature"}},
                        {"type": "EnterFeature", "data": {"feature": "test_feature"}},
                        {
                            "type": "StickyAssignment",
                            "data": {
                                "variant": None,
                                "value": None,
                                "assigned": False,
                                "ts": None,
                                "source": None,
                            },
                        },
                        {"type": "EnterRollout", "data": {"rollout": "default"}},
                        {
                            "type": "EvaluatePopulation",
                            "data": {
                                "member": True,
                                "population": "Default",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "_Field",
                                "args": [{"id": "two"}, "$id"],
                                "result": "two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Concat",
                                "args": ["default", ":", "two"],
                                "result": "default:two",
                            },
                        },
                        {
                            "type": "EvalFunc",
                            "data": {
                                "f": "Hash",
                                "args": ["default:two"],
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "Randomize",
                            "data": {
                                "function": "Hash(Concat('default', ':', $id))",
                                "result": 0.7560122309797355,
                            },
                        },
                        {
                            "type": "EnterArm",
                            "data": {
                                "arm": {"type": "Arm", "variant": "foo", "weight": 1.0},
                                "cutoff": 1.0,
                                "x": 0.7560122309797355,
                            },
                        },
                        {"type": "LeaveArm", "data": {"matched": True}},
                        {"type": "LeaveRollout", "data": {"member": True}},
                        {
                            "type": "ChoseVariant",
                            "data": {"variant": "foo", "sticky": False},
                        },
                        {"type": "EnterVariant", "data": {"variant": "foo"}},
                        {"type": "LeaveVariant", "data": {"value": "Foo"}},
                        {"type": "LeaveFeature", "data": {"value": "Foo"}},
                        {"type": "LeaveGate", "data": {"value": "Foo"}},
                    ],
                }
            ]
        )


class TestNetworkLogger(unittest.IsolatedAsyncioTestCase):
    @responses.activate
    def test_log_simple(self):
        responses.add(responses.POST, mock_url, json={"data": "ok"})

        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        logger = NetworkLogger(
            mock_url, headers={"Authorization": "Bearer glen"}, debug=True
        )
        v = asyncio.run(f(User("one"), log=logger, now=mock_now))
        v.log()

        wait_for_responses(logger)

        assert len(responses.calls) == 1
        assert responses.calls[0].request.headers.get("Authorization") == "Bearer glen"
        assert (
            responses.calls[0].request.headers.get("Content-Type")
            == "application/json; charset=utf-8"
        )
        assert json.loads(responses.calls[0].request.body) == {
            "ts": "2022-01-29T12:11:10+00:00",
            "call_id": "e3e70682-c209-4cac-629f-6fbed82c07cd",
            "entity": {
                "type": "User",
                "value": {
                    "id": "one",
                },
            },
            "feature": {
                "name": "test_feature",
                "rollouts": [
                    {
                        "arms": [
                            {
                                "type": "Arm",
                                "variant": "foo",
                                "weight": 1.0,
                            }
                        ],
                        "name": "default",
                        "population": {
                            "name": "Default",
                            "type": "Population",
                        },
                        "randomizer": "Hash(Concat('default', ':', $id))",
                        "sticky": None,
                        "type": "Rollout",
                    }
                ],
                "type": "Feature",
                "variants": {
                    "foo": {
                        "name": "foo",
                        "nested": False,
                        "type": "Variant",
                        "value": "Foo",
                    },
                },
            },
            "variant": {
                "name": "foo",
                "nested": False,
                "type": "Variant",
                "value": "Foo",
            },
            "assignment": "Foo",
            "trace": None,
            "repeat": False,
            "sticky": False,
        }

    @responses.activate
    def test_log_retry(self):
        responses.add(responses.POST, mock_url, status=500)
        responses.add(responses.POST, mock_url, status=500)
        responses.add(responses.POST, mock_url, json={"data": {"id": "mockuid"}})
        logger = NetworkLogger(
            mock_url,
            headers={"Authorization": "Bearer glen"},
            body=lambda x: {"custom": "foo"},
            debug=True,
        )

        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        v = asyncio.run(f(User("one"), log=logger, now=mock_now))
        v.log()

        wait_for_responses(logger)

        assert len(responses.calls) == 3
        assert responses.calls[0].request.body == '{"custom": "foo"}'

    @responses.activate
    def test_log_format(self):
        responses.add(responses.POST, mock_url, json={"data": {"id": "mockuid"}})

        logger = NetworkLogger(
            mock_url,
            headers={"Authorization": "Bearer glen"},
            body=lambda x: {"custom": "foo"},
            debug=True,
        )

        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        v = asyncio.run(f(User("one"), log=logger, now=mock_now))
        v.log()

        wait_for_responses(logger)

        assert len(responses.calls) == 1
        assert responses.calls[0].request.body == '{"custom": "foo"}'

    @responses.activate
    def test_log_skip(self):
        def skip_log(x):
            raise SkipLog

        logger = NetworkLogger(
            mock_url,
            headers={"Authorization": "Bearer glen"},
            body=skip_log,
            debug=True,
        )

        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        v = asyncio.run(f(User("one"), log=logger, now=mock_now))
        v.log()

        wait_for_responses(logger, expected_count=0)

        assert len(responses.calls) == 0



import unittest
from datetime import UTC, datetime

import crocodsl.func as func


class TestFunc(unittest.TestCase):
    def test_op_repr(self):
        """Test the representations of the operators."""
        assert repr(func.Eq) == "Eq"
        assert repr(func.Ne) == "Not(Eq)"
        assert repr(func.And) == "And"
        assert repr(func.Or) == "Or"
        assert repr(func.Not) == "Not"
        assert repr(func.Lt) == "Lt"
        assert repr(func.Le) == "Le"
        assert repr(func.Gt) == "Not(Le)"
        assert repr(func.Ge) == "Not(Lt)"
        assert repr(func.In) == "In"
        assert repr(func.Concat) == "Concat"
        assert repr(func.Hash) == "Hash"
        assert repr(func.Len) == "Len"
        assert repr(func.Matches) == "Matches"
        assert repr(func.Now) == "Now"
        assert repr(func.TimeSince) == "TimeSince"

    def test_expr_repr(self):
        """Test the representations of expressions."""
        assert repr(func.Eq("a", "b")) == "'a' Eq 'b'"
        assert repr(func.Ne("a", "b")) == "Not('a' Eq 'b')"
        assert (
            repr(
                func.And(
                    func.Eq("a", "a"),
                    func.Lt(10, 11),
                )
            )
            == "('a' Eq 'a') And (10 Lt 11)"
        )
        assert (
            repr(
                func.Or(
                    func.Eq("a", "b"),
                    func.Lt(10, 11),
                )
            )
            == "('a' Eq 'b') Or (10 Lt 11)"
        )
        assert repr(func.Not(True)) == "Not(True)"
        assert repr(func.Le(10, 10)) == "10 Le 10"
        assert repr(func.Gt(11, 10)) == "Not(11 Le 10)"
        assert repr(func.Ge(11, 10)) == "Not(11 Lt 10)"
        assert repr(func.In("a", ["a", "b", "c"])) == "'a' In ['a', 'b', 'c']"
        assert repr(func.Concat("a", "b", "c")) == "Concat('a', 'b', 'c')"
        assert repr(func.Hash("foo")) == "Hash('foo')"
        assert repr(func.Has([1, 2, 3], 1) == "1 In [1, 2, 3]")
        assert repr(func.Matches("Foo", r".*") == "'Foo' Matches '.*'")

    def test_eq(self):
        """Test equality."""
        assert func.Eq("a", "a")() is True
        assert func.Eq("a", "b")() is False
        assert func.Eq(1, 1)() is True
        assert func.Eq(1, 0)() is False

    def test_ne(self):
        """Test inequality."""
        assert func.Ne("a", "a")() is False
        assert func.Ne("a", "b")() is True
        assert func.Ne(1, 1)() is False
        assert func.Ne(1, 0)() is True

    def test_and(self):
        """Test and."""
        assert (
            func.And(
                func.Eq("a", "a"),
                func.Eq(1, 1),
            )()
            is True
        )
        assert (
            func.And(
                func.Eq("a", "a"),
                func.Eq(1, 0),
            )()
            is False
        )
        assert (
            func.And(
                func.Eq("a", "b"),
                func.Eq(1, 0),
            )()
            is False
        )
        assert (
            func.And(
                func.Eq("a", "b"),
                func.Eq(1, 1),
            )()
            is False
        )

    def test_or(self):
        """Test or."""
        assert (
            func.Or(
                func.Eq("a", "a"),
                func.Eq(1, 1),
            )()
            is True
        )
        assert (
            func.Or(
                func.Eq("a", "a"),
                func.Eq(1, 0),
            )()
            is True
        )
        assert (
            func.Or(
                func.Eq("a", "b"),
                func.Eq(1, 0),
            )()
            is False
        )
        assert (
            func.Or(
                func.Eq("a", "b"),
                func.Eq(1, 1),
            )()
            is True
        )

    def test_not(self):
        """Test negation."""
        assert func.Not(True)() is False
        assert func.Not(False)() is True

    def test_lt(self):
        """Test < equality."""
        assert func.Lt(10, 11)() is True
        assert func.Lt(11, 11)() is False
        assert func.Lt(12, 11)() is False

    def test_le(self):
        """Test <= equality."""
        assert func.Le(10, 11)() is True
        assert func.Le(11, 11)() is True
        assert func.Le(12, 11)() is False

    def test_gt(self):
        """Test > equality."""
        assert func.Gt(10, 11)() is False
        assert func.Gt(11, 11)() is False
        assert func.Gt(12, 11)() is True

    def test_ge(self):
        """Test > equality."""
        assert func.Ge(10, 11)() is False
        assert func.Ge(11, 11)() is True
        assert func.Ge(12, 11)() is True

    def test_in(self):
        """Test containment."""
        assert func.In("a", ["a", "b", "c"])() is True
        assert func.In("b", ["a", "b", "c"])() is True
        assert func.In("c", ["a", "b", "c"])() is True
        assert func.In("x", ["a", "b", "c"])() is False
        assert func.In("x", None)() is False

    def test_has(self):
        """Test reverse containment."""
        assert func.Has(["a", "b", "c"], "a")() is True
        assert func.Has(["a", "b", "c"], "b")() is True
        assert func.Has(["a", "b", "c"], "c")() is True
        assert func.Has(["a", "b", "c"], "x")() is False
        assert func.Has(None, "x")() is False

    def test_concat(self):
        """Test concatenation."""
        assert func.Concat("a", "b", "c")() == "abc"

    def test_hash(self):
        """Test hashing."""
        assert func.Hash("abc")() == 0.7054175881782409

    def test_len(self):
        """Test length."""
        assert func.Len([1, 2, 3]) == 3
        assert func.Len("abc") == 3
        assert (
            func.Len(
                (
                    "a",
                    "b",
                )
            )
            == 2
        )
        assert func.Len({"a": 1, "b": 2, "c": 3}) == 3

    def test_matches(self):
        """Test regex matching."""
        assert func.Matches("foo", r".*")() is True
        assert func.Matches("foo", r"^f")() is True
        assert func.Matches("bar", r"^f")() is False
        assert func.Matches("a.b.c", r"\.b\.")() is True

    def test_now(self):
        """Test now function."""
        ts = datetime(2024, 1, 2, 3, 4, 5, tzinfo=UTC)
        assert func.Now()(context={"now": lambda: ts}) == ts

    def test_time_since(self):
        """Test time delta computation."""
        orig_ts = datetime(2024, 1, 1, 3, 4, 5, tzinfo=UTC)
        ts = datetime(2024, 1, 2, 3, 4, 5, tzinfo=UTC)
        assert func.TimeSince(orig_ts, "seconds")(context={"now": lambda: ts}) == 86_400
        assert func.TimeSince(orig_ts, "minutes")(context={"now": lambda: ts}) == 1_440
        assert func.TimeSince(orig_ts, "hours")(context={"now": lambda: ts}) == 24
        assert func.TimeSince(orig_ts, "d")(context={"now": lambda: ts}) == 1
        assert func.TimeSince(orig_ts, "w")(context={"now": lambda: ts}) == 1 / 7
        assert (
            func.TimeSince(orig_ts, "mo")(context={"now": lambda: ts})
            == 86_400 / 2_628_288
        )
        assert func.TimeSince(orig_ts, "y")(context={"now": lambda: ts}) == 1 / 365

    def test_trim_prefix(self):
        """Test removal of characters from a string."""
        assert repr(func.TrimPrefix("abc", "b")) == "TrimPrefix('abc', 'b')"
        assert func.TrimPrefix("abc", "ab")() == "c"
        assert func.TrimPrefix("abc", "a")() == "bc"
        assert func.TrimPrefix("abc", "c")() == "abc"
        assert func.TrimPrefix("prefixabc", "pref")() == "ixabc"
        with self.assertRaises(TypeError):
            func.TrimPrefix(None, "a")()
        with self.assertRaises(TypeError):
            func.TrimPrefix("abc", None)()

    def test_trim_suffix(self):
        """Test removal of characters from a string."""
        assert repr(func.TrimSuffix("abc", "b")) == "TrimSuffix('abc', 'b')"
        assert func.TrimSuffix("abc", "bc")() == "a"
        assert func.TrimSuffix("abc", "c")() == "ab"
        assert func.TrimSuffix("abc", "a")() == "abc"
        assert func.TrimSuffix("abcsuffix", "suffix")() == "abc"
        with self.assertRaises(TypeError):
            func.TrimSuffix(None, "a")()
        with self.assertRaises(TypeError):
            func.TrimSuffix("abc", None)()

    def test_composition(self):
        """Make sure functions can be composed."""
        assert func.Hash(func.Concat("a", "b", "c"))() == 0.7054175881782409

    def test_dunders(self):
        """Test that expressions can be combined algebraically."""
        assert repr(func.Hash("abc") < 0.75) == "Hash('abc') Lt 0.75"
        assert (func.Hash("abc") < 0.75)() is True

        assert repr(func.Hash("abc") <= 0.75) == "Hash('abc') Le 0.75"
        assert (func.Hash("abc") <= 0.75)() is True

        assert repr(func.Hash("abc") > 0.7) == "Not(Hash('abc') Le 0.7)"
        assert (func.Hash("abc") > 0.7)() is True

        assert repr(func.Hash("abc") >= 0.7) == "Not(Hash('abc') Lt 0.7)"
        assert (func.Hash("abc") >= 0.7)() is True

        assert (
            repr(func.Hash("abc") == 0.7054175881782409)
            == "Hash('abc') Eq 0.7054175881782409"
        )
        assert (func.Hash("abc") == 0.7054175881782409)() is True

        assert repr(func.Hash("abc") != 0.7) == "Not(Hash('abc') Eq 0.7)"
        assert (func.Hash("abc") != 0.7)() is True

        assert repr(func.Lt(10, 11).and_(func.Lt(1, 2))) == "(10 Lt 11) And (1 Lt 2)"
        assert func.Lt(10, 11).and_(func.Lt(1, 2))() is True

        assert repr(func.Lt(10, 11).or_(func.Lt(2, 1))) == "(10 Lt 11) Or (2 Lt 1)"
        assert func.Lt(10, 11).or_(func.Lt(2, 1))() is True

        assert repr(func.Literal([1, 2, 3]).has(1)) == "1 In [1, 2, 3]"
        assert func.Literal([1, 2, 3]).has(1)() is True



import re
from collections.abc import Iterable, Sequence

from .common import hash_id, utcnow


class _MetaExpression(type):
    """Metaclass for all expressions."""

    def __getitem__(self, other):
        """Compose one function with another, in the form f(g(x)).

        This enables syntactic sugar, for example defining the "not equals"
        object as:
        Ne = Not[Eq]

        Args:
            other - function to compose

        Returns:
            Composed expression.
        """
        return _ComposedExpression(self, other)

    def __repr__(self) -> str:
        return self.__name__


class _Expression(metaclass=_MetaExpression):
    """Base class for an arbitrary expression.

    This implements boolean algebra on expressions; expression can be compared
    to each other with the standard equality or ordering operations.
    """

    def __eq__(self, other):
        return Eq(self, other)

    def __ne__(self, other):
        return Ne(self, other)

    def __lt__(self, other):
        return Lt(self, other)

    def __le__(self, other):
        return Le(self, other)

    def __gt__(self, other):
        return Gt(self, other)

    def __ge__(self, other):
        return Ge(self, other)

    def in_(self, other):
        return In(self, other)

    def or_(self, other):
        return Or(self, other)

    def and_(self, other):
        return And(self, other)

    def has(self, other):
        # Note __contains__ coerces to bool so we can't use it here.
        return Has(self, other)

    def matches(self, other):
        return Matches(self, other)

    def validate(self):
        """Validation for an expression can be implemented in a subclass."""
        pass

    def __call__(self, *args, log=None, context=None):
        """This is the actual behavior of the operator.

        This must be implemented in a subclass."""

        raise NotImplementedError()

    def equivalent(self, other):
        """Check whether two expressions are equivalent.

        Args:
            other - another _Expression

        Returns:
            True if expressions are the same.
        """
        if not isinstance(other, _Expression):
            return False

        return str(self) == str(other)

    def _trace(self, log, args, result):
        """Emit a EvalFunc trace event. If there is no logger passed, this is
        a no-op.

        Args:
            log - Log function (can be None)
            args - List of (evaluated) arguments used for evaluating this
            expression.
            result - Evaluation result
        """
        if not log:
            return
        name = self.__class__.__name__
        log(name, args, result)

    def to_json(self):
        return str(self)


class _ComposedExpression(_Expression):
    """Composition operator, as g(f(x))."""

    def __init__(self, g, f):
        self.outer = g
        if isinstance(f, Iterable):
            self.inners = f
        else:
            self.inners = (f,)

    def __call__(self, *args, **kwargs):
        # kwargs are contextual things like `log` that should be shared across
        # all invocations.
        inner_args = [f(*args, **kwargs) for f in self.inners]
        return self.outer(*inner_args, **kwargs)

    def __repr__(self):
        all_reps = [repr(inner) for inner in self.inners]
        if len(self.inners) == 1:
            inners = all_reps[0]
        else:
            inners = ", ".join(all_reps)
        return f"{repr(self.outer)}({inners})"


class _NullaryExpression(_Expression):
    """Expression of the form `operator()`."""

    def __init__(self):
        pass

    def evaluate(self, *args, log=None, context=None):
        pass


class _UnaryExpression(_Expression):
    """Expression of the form `operator(a)`."""

    def __init__(self, arg):
        self.arg = arg

    def evaluate(self, *args, log=None, context=None):
        arg = self.arg

        if callable(arg):
            arg = arg(*args, log=log, context=context)

        return arg

    def __repr__(self):
        op = repr(self.__class__)
        return f"{op}({repr(self.arg)})"


class _BinaryExpression(_Expression):
    """Expression of the form `operator(a, b)`."""

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def evaluate(self, *args, log=None, context=None):
        left = self.left
        right = self.right

        if callable(left):
            left = left(*args, log=log, context=context)

        if callable(right):
            right = right(*args, log=log, context=context)

        return left, right

    def __repr__(self):
        op = repr(self.__class__)
        return f"{op}({repr(self.left)}, {repr(self.right)})"


class _InfixExpression(_BinaryExpression):
    """BinaryExpression with the representation `a <operator> b`."""

    def __repr__(self):
        op = repr(self.__class__)

        # Apply parentheses to sub-expressions if they are infix expressions.
        # Other notations are not ambiguous.
        if isinstance(self.left, _InfixExpression):
            left = f"({repr(self.left)})"
        else:
            left = repr(self.left)

        if isinstance(self.right, _InfixExpression):
            right = f"({repr(self.right)})"
        else:
            right = repr(self.right)

        return f"{left} {op} {right}"


class _NAryExpression(_Expression):
    def __init__(self, *args):
        self.args = args

    def evaluate(self, *fargs, log=None, context=None):
        return [
            a(*fargs, log=log, context=context) if callable(a) else a for a in self.args
        ]

    def __repr__(self):
        op = repr(self.__class__)
        args = [repr(a) for a in self.args]
        return f"{op}({', '.join(args)})"


# Operator definitions
#
# These are all the operators available to expressions.


class Literal(_UnaryExpression):
    """A literal value."""

    def __call__(self, *args, log=None, context=None):
        arg = self.evaluate(*args, log=log, context=context)

        self._trace(log, [arg], arg)

        return arg

    def __repr__(self):
        return repr(self.arg)


class Or(_InfixExpression):
    """Boolean 'or' operator."""

    def __call__(self, *args, log=None, context=None):
        left, right = self.evaluate(*args, log=log, context=context)
        result = left or right

        self._trace(log, [left, right], result)

        return result


class And(_InfixExpression):
    """Boolean 'and' operator."""

    def __call__(self, *args, log=None, context=None):
        left, right = self.evaluate(*args, log=log, context=context)
        result = left and right

        self._trace(log, [left, right], result)

        return result


class Not(_UnaryExpression):
    """Negate a value."""

    def __call__(self, *args, log=None, context=None):
        arg = self.evaluate(*args, log=log, context=context)
        result = not arg

        self._trace(log, [arg], result)

        return result


class Eq(_InfixExpression):
    """Equals operator."""

    def __call__(self, *args, log=None, context=None):
        left, right = self.evaluate(*args, log=log, context=context)
        result = left == right

        self._trace(log, [left, right], result)

        return result


Ne = Not[Eq]  # type: ignore
"""Not equals operator."""


class Lt(_InfixExpression):
    """Less than operator."""

    def __call__(self, *args, log=None, context=None):
        left, right = self.evaluate(*args, log=log, context=context)
        result = left < right

        self._trace(log, [left, right], result)

        return result


class Le(_InfixExpression):
    """Less than or equal to operator."""

    def __call__(self, *args, log=None, context=None):
        left, right = self.evaluate(*args, log=log, context=context)
        result = left <= right

        self._trace(log, [left, right], result)

        return result


Gt = Not[Le]  # type: ignore
"""Greater than operator."""


Ge = Not[Lt]  # type: ignore
"""Greater than or equal to operator."""


class In(_InfixExpression):
    """Containment operator

    If the left side is a collection, this returns True if *any* of the left
    side are in the right.
    """

    def __call__(self, *args, log=None, context=None):
        left, right = self.evaluate(*args, log=log, context=context)

        # Treat `None` as an empty list
        if right is None:
            right = []

        result = False
        if isinstance(left, Sequence) and not isinstance(left, str):
            result = any((x in right for x in left))
        else:
            result = left in right

        self._trace(log, [left, right], result)

        return result


def Has(left, right):
    """Reversed notation of `In`."""
    return In(right, left)


class Concat(_NAryExpression):
    """Concatenation operator."""

    def __call__(self, *args, log=None, context=None):
        args = self.evaluate(*args, log=log, context=context)
        val = "".join([str(a) for a in args])

        self._trace(log, args, val)

        return val


class Hash(_UnaryExpression):
    """Compute the assignment hash of a value."""

    def __call__(self, *args, log=None, context=None):
        arg = self.evaluate(*args, log=log, context=context)
        val = hash_id(str(arg))

        self._trace(log, [arg], val)

        return val


class Len(_UnaryExpression):
    """Get length of an iterable."""

    def __call__(self, *args, log=None, context=None):
        arg = self.evaluate(*args, log=log, context=context)
        val = len(arg)

        self._trace(log, [arg], val)

        return val


class Matches(_InfixExpression):
    """Perform a regular expression match.

    Example:
        $my_prop Matches '.*'
    """

    def __call__(self, *args, log=None, context=None):
        left, right = self.evaluate(*args, log=log, context=context)

        if left is None:
            left = ""
        if right is None:
            right = ""

        result = re.search(right, left) is not None

        self._trace(log, [left, right], result)

        return result


class Now(_NullaryExpression):
    """Get current time."""

    def __call__(self, *args, log=None, context=None):
        now_fn = utcnow

        if context and "now" in context:
            now_fn = context["now"]

        result = now_fn()

        self._trace(log, [], result)

        return result


class TimeSince(_BinaryExpression):
    """Get amount of time elapsed since a moment in the given unit.

    Example:
        TimeSince($my_prop, 'months')
    """

    _CONVERSIONS = {
        unit: conversion
        for units, conversion in [
            ({"s", "sec", "second", "seconds"}, 1.0),
            ({"m", "min", "minute", "minutes"}, 60.0),
            ({"h", "hr", "hour", "hours"}, 3_600.0),
            ({"d", "day", "days"}, 86_400.0),
            ({"w", "wk", "week", "weeks"}, 604_800.0),
            ({"mo", "mon", "month", "months"}, 2_628_288.0),
            ({"y", "yr", "year", "years"}, 31_536_000.0),
        ]
        for unit in units
    }

    def __call__(self, *args, log=None, context=None):
        now_ts = Now()(*args, log=log, context=context)
        moment, unit = self.evaluate(*args, log=log, context=context)

        delta = now_ts - moment
        seconds = delta.total_seconds()
        result = self._convert(seconds, unit)

        self._trace(log, [now_ts, moment, unit], result)
        return result

    def _convert(self, seconds, unit):
        unit = unit.lower()
        try:
            conversion = self._CONVERSIONS[unit.lower()]
            return seconds / conversion
        except KeyError:
            raise ValueError(f"Invalid unit '{unit}'")


class TrimPrefix(_BinaryExpression):
    """Remove a prefix from a string.

    Example:
        TrimPrefix($my_prop, 'prefix')
    """

    def __call__(self, *args, log=None, context=None):
        s, prefix = self.evaluate(*args, log=log, context=context)

        if not isinstance(s, str):
            raise TypeError(f"Expect {type(s)} to be str")

        result = s.removeprefix(prefix)

        self._trace(log, [s, prefix], result)

        return result


class TrimSuffix(_BinaryExpression):
    """Remove a suffix from a string.

    Example:
        TrimSuffix($my_prop, 'suffix')
    """

    def __call__(self, *args, log=None, context=None):
        s, suffix = self.evaluate(*args, log=log, context=context)

        if not isinstance(s, str):
            raise TypeError(f"Expect {type(s)} to be str")

        result = s.removesuffix(suffix)

        self._trace(log, [s, suffix], result)

        return result



import abc
from typing import TYPE_CHECKING, Any, Optional

import alligater.events as events
import crocodsl.field as field
import crocodsl.func as func

from .common import ValidationError, NowFn, default_now

if TYPE_CHECKING:
    from . import Alligater


class PopulationSelector(abc.ABC):
    @abc.abstractmethod
    def validate(self): ...

    @abc.abstractmethod
    async def __call__(
        self,
        call_id: str,
        entity: Any,
        log: Optional[events.EventLogger],
        gater: Optional["Alligater"],
        now: NowFn = default_now,
    ) -> bool: ...

    @abc.abstractmethod
    def to_dict(self) -> dict: ...


class DefaultSelector(PopulationSelector):
    """Selector for 100% of the population."""

    def validate(self):
        """Ensures configuration is correct. (It is.)"""
        pass

    async def __call__(self, call_id, entity, log=None, gater=None, now=default_now):
        """Returns True -- everyone is in the default population!

        Args:
            call_id - ID of the exposure evaluation call
            entity - Entity to test

        Returns:
            Always True
        """
        events.EvaluatePopulation(
            log,
            population=self,
            entity=entity,
            member=True,
            call_id=call_id,
            now=now,
        )
        return True

    def __eq__(self, other):
        if not isinstance(other, DefaultSelector):
            return False

        return True

    def __repr__(self):
        return "<Population default>"

    def to_dict(self):
        return {
            "type": "Population",
            "name": "Default",
        }


class ExpressionSelector(PopulationSelector):
    """Selector for a segment of a population based on the given Expression."""

    def __init__(self, expression):
        """Create a selector with the given Expression.

        Args:
            expression - Expression to evaluate to test membership
        """
        self.expression = expression

    def validate(self):
        """Ensures configuration makes sense.

        Raises:
            ValidationError - when something is wrong
        """
        self.expression.validate()

    async def __call__(self, call_id, entity, log=None, gater=None, now=default_now):
        """Check whether entity belongs to a this population.

        Args:
            call_id - ID of the exposure evaluation call
            entity - The entity to test

        Returns:
            True or False indicating membership in this population.
        """

        def trace(name, args, result):
            events.EvalFunc(
                log, f=name, args=args, result=result, call_id=call_id, now=now
            )

        result = self.expression(entity, log=trace, context={"now": now})
        events.EvaluatePopulation(
            log, population=self, entity=entity, member=result, call_id=call_id, now=now
        )
        return result

    def __eq__(self, other):
        if not isinstance(other, ExpressionSelector):
            return False

        return self.expression.equivalent(other.expression)

    def __repr__(self):
        return "<Population expression=({})>".format(self.expression)

    def to_dict(self):
        return {
            "type": "Population",
            "name": "Expression",
            "expression": str(self.expression),
        }


class FeatureSelector(ExpressionSelector):
    """Select a population based on a feature of the entity."""

    def __init__(self, feature, expr):
        """Select a population based on another feature.

        Args:
            feature - Feature to delegate to
            expr - Expression on feature evaluation. $variant, $value, and $assigned are
            variables available for conditioning.
        """
        self.feature = feature
        super().__init__(expr)

    async def __call__(self, call_id, entity, log=None, gater=None, now=default_now):
        """Check whether entity belongs to a this population.

        Args:
            call_id - ID of the exposure evaluation call
            entity - The entity to test

        Returns:
            True or False indicating membership in this population.
        """
        # NOTE: only assignments get logged from this sub-evaluation, never
        # any exposures. In theory the parent exposure should be logged in
        # lieu of the child.
        value = await getattr(gater, self.feature)(entity, deferred=True, now=now)
        result_entity = {
            "value": value.value,
            "variant": value.variant,
            "assigned": value.ts,
        }
        return await super().__call__(
            call_id, result_entity, log=log, gater=gater, now=now
        )


class PercentSelector(ExpressionSelector):
    """Select a percentage of the population."""

    def __init__(self, percent, seed, id_field=field.ID):
        """Select a percentage of the population.

        Args:
            percent - Value in [0, 1] representing % to select.
            seed - Seed to use for randomization. This functions as the salt
            of the hash for reproducible assignments.
            id_field - Field of the Entity to use for hashing.
        """
        self.percent = percent
        expression = func.Hash(func.Concat(seed, id_field)) <= percent
        super().__init__(expression)

    def validate(self):
        """Ensures configuration makes sense.

        Raises:
            ValidationError - when config doesn't make sense
        """
        if self.percent < 0.0:
            raise ValidationError("PercentSelector has to have a positive value")

        if self.percent > 1.0:
            raise ValidationError("PercentSelector has to be less than 1.0")


class ExplicitSelector(ExpressionSelector):
    """Select specific individuals from the population."""

    def __init__(self, ids, id_field=field.ID):
        """Select an explicit set of IDs.

        Args:
            ids - iterable of IDs
            id_field - Field to use as the ID field of the entity
        """
        self.ids = ids
        expression = id_field.in_(ids)
        super().__init__(expression)


class Population:
    """Predefined populations."""

    DEFAULT = DefaultSelector()
    Percent = PercentSelector
    Expression = ExpressionSelector
    Feature = FeatureSelector
    Explicit = ExplicitSelector
